@@ -37,12 +37,16 @@ CONF = cfg.CONF
 
 BIND_MOUNTS = ('/dev', '/proc', '/run')
 
+# NOTE(TheJulia): Do not add bootia32.csv to this list. That is 32bit
+# EFI booting and never really became popular.
 BOOTLOADERS_EFI = [
+    'bootx64.csv',  # Used by GRUB2 shim loader (Ubuntu, Red Hat)
+    'boot.csv',  # Used by rEFInd, Centos7 Grub2
     'bootia32.efi',
-    'bootx64.efi',
+    'bootx64.efi',  # x86_64 Default
     'bootia64.efi',
     'bootarm.efi',
-    'bootaa64.efi',
+    'bootaa64.efi',  # Arm64 Default
     'bootriscv32.efi',
     'bootriscv64.efi',
     'bootriscv128.efi',
@@ -223,9 +227,10 @@ def _is_bootloader_loaded(dev):
 def _get_efi_bootloaders(location):
     """Get all valid efi bootloaders in a given location
 
-    :param location: the location where it should  start looking for the
+    :param location: the location where it should start looking for the
                      efi files.
-    :return: a list of relative paths to valid efi bootloaders
+    :return: a list of relative paths to valid efi bootloaders or reference
+             files.
     """
     # Let's find all files with .efi or .EFI extension
     LOG.debug('Looking for all efi files on %s', location)
@@ -241,19 +246,34 @@ def _get_efi_bootloaders(location):
                 v_bl = efi_f.split(location)[-1][1:]
                 LOG.debug('%s is a valid bootloader', v_bl)
                 valid_bootloaders.append(v_bl)
+            if 'csv' in efi_f.lower():
+                v_bl = efi_f.split(location)[-1][1:]
+                LOG.debug('%s is a pointer to a bootloader', v_bl)
+                # The CSV files are intended to be authortative as
+                # to the bootloader and the label to be used. Since
+                # we found one, we're going to point directly to it.
+                # centos7 did ship with 2, but with the same contents.
+                # TODO(TheJulia): Perhaps we extend this to make a list
+                # of CSVs instead and only return those?! But then the
+                # question is which is right/first/preferred.
+                return [v_bl]
     return valid_bootloaders
 
 
-def _run_efibootmgr(valid_efi_bootloaders, device, efi_partition):
+def _run_efibootmgr(valid_efi_bootloaders, device, efi_partition,
+                    mount_point):
     """Executes efibootmgr and removes duplicate entries.
 
     :param valid_efi_bootloaders: the list of valid efi bootloaders
     :param device: the device to be used
     :param efi_partition: the efi partition on the device
+    :param mount_point: The mountpoint for the EFI partition so we can
+                        read contents of files if necessary to perform
+                        proper bootloader injection operations.
     """
 
     # Before updating let's get information about the bootorder
-    LOG.debug("Getting information about boot order")
+    LOG.debug("Getting information about boot order.")
     utils.execute('efibootmgr')
     # NOTE(iurygregory): regex used to identify the Warning in the stderr after
     # we add the new entry. Example:
@@ -262,13 +282,27 @@ def _run_efibootmgr(valid_efi_bootloaders, device, efi_partition):
                                   r'Boot([0-9a-f-A-F]+)\s.*$')
     label_id = 1
     for v_bl in valid_efi_bootloaders:
-        v_efi_bl_path = '\\' + v_bl.replace('/', '\\')
-        # Update the nvram using efibootmgr
-        # https://linux.die.net/man/8/efibootmgr
-        label = 'ironic' + str(label_id)
+        if 'csv' in v_bl.lower():
+            LOG.debug('A CSV file has been identified as a bootloader hint. '
+                      'File: %s', v_bl)
+            # These files are always UTF-16 encoded, sometimes have a header.
+            # Positive bonus is python silently drops the FEFF header.
+            with open(mount_point + '/' + v_bl, 'r', encoding='utf-16') as csv:
+                contents = str(csv.read())
+            csv_contents = contents.split(',', maxsplit=3)
+            csv_filename = v_bl.split('/')[-1]
+            v_efi_bl_path = v_bl.replace(csv_filename, str(csv_contents[0]))
+            v_efi_bl_path = '\\' + v_efi_bl_path.replace('/', '\\')
+            label = csv_contents[1]
+        else:
+            v_efi_bl_path = '\\' + v_bl.replace('/', '\\')
+            label = 'ironic' + str(label_id)
+
         LOG.debug("Adding loader %(path)s on partition %(part)s of device "
                   " %(dev)s", {'path': v_efi_bl_path, 'part': efi_partition,
                                'dev': device})
+        # Update the nvram using efibootmgr
+        # https://linux.die.net/man/8/efibootmgr
         cmd = utils.execute('efibootmgr', '-c', '-d', device,
                             '-p', efi_partition, '-w', '-L', label,
                             '-l', v_efi_bl_path)
@@ -296,7 +330,7 @@ def _manage_uefi(device, efi_system_part_uuid=None):
     """
     efi_partition_mount_point = None
     efi_mounted = False
-
+    LOG.debug('Attempting UEFI loader autodetection and NVRAM record setup.')
     try:
         # Force UEFI to rescan the device.
         _rescan_device(device)
@@ -343,11 +377,13 @@ def _manage_uefi(device, efi_system_part_uuid=None):
 
         valid_efi_bootloaders = _get_efi_bootloaders(efi_partition_mount_point)
         if valid_efi_bootloaders:
-            _run_efibootmgr(valid_efi_bootloaders, device, efi_partition)
+            _run_efibootmgr(valid_efi_bootloaders, device, efi_partition,
+                            efi_partition_mount_point)
             return True
         else:
             # NOTE(dtantsur): if we have an empty EFI partition, try to use
             # grub-install to populate it.
+            LOG.warning('Empty EFI partition detected.')
             return False
 
     except processutils.ProcessExecutionError as e:
@@ -518,6 +554,29 @@ def _umount_all_partitions(path, path_variable, umount_warn_msg):
     return umount_binds_success
 
 
+def _mount_partition(partition, path):
+    if not os.path.ismount(path):
+        LOG.debug('Attempting to mount %(device)s to %(path)s to '
+                  'partition.',
+                  {'device': partition,
+                   'path': path})
+        try:
+            utils.execute('mount', partition, path)
+        except processutils.ProcessExecutionError as e:
+            # NOTE(TheJulia): It seems in some cases,
+            # the python os.path.ismount can return False
+            # even *if* it is actually mounted. This appears
+            # to be becasue it tries to rely on inode on device
+            # logic, yet the rules are sometimes different inside
+            # ramdisks. So lets check the error first.
+            if 'already mounted' not in e:
+                # Raise the error, since this is not a known
+                # failure case
+                raise
+            else:
+                LOG.debug('Partition already mounted, proceeding.')
+
+
 def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
                    prep_boot_part_uuid=None, target_boot_mode='bios'):
     """Install GRUB2 bootloader on a given device."""
@@ -605,11 +664,9 @@ def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
                 # remounted.
                 LOG.debug('No EFI assets were preserved for setup or the '
                           'ramdisk was unable to complete the setup. '
-                          'falling back to bootloader installation from'
+                          'falling back to bootloader installation from '
                           'deployed image.')
-                if not os.path.ismount(path):
-                    LOG.debug('Re-mounting the root partition.')
-                    utils.execute('mount', root_partition, path)
+                _mount_partition(root_partition, path)
 
         binary_name = "grub"
         if os.path.exists(os.path.join(path, 'usr/sbin/grub2-install')):
@@ -629,36 +686,41 @@ def _install_grub2(device, root_uuid, efi_system_part_uuid=None,
             LOG.warning("GRUB2 will be installed for UEFI on efi partition "
                         "%s using the install command which does not place "
                         "Secure Boot signed binaries.", efi_partition)
-            utils.execute('mount', efi_partition, efi_partition_mount_point)
+
+            _mount_partition(efi_partition, efi_partition_mount_point)
             efi_mounted = True
-            # FIXME(rg): does not work in cross boot mode case (target
-            # boot mode differs from ramdisk one)
-            # Probe for the correct target (depends on the arch, example
-            # --target=x86_64-efi)
-            utils.execute('chroot %(path)s /bin/sh -c '
-                          '"%(bin)s-install"' %
-                          {'path': path, 'bin': binary_name},
-                          shell=True,
-                          env_variables={
-                              'PATH': path_variable
-                          })
-            # Also run grub-install with --removable, this installs grub to
-            # the EFI fallback path. Useful if the NVRAM wasn't written
-            # correctly, was reset or if testing with virt as libvirt
-            # resets the NVRAM on instance start.
-            # This operation is essentially a copy operation. Use of the
-            # --removable flag, per the grub-install source code changes
-            # the default file to be copied, destination file name, and
-            # prevents NVRAM from being updated.
-            # We only run grub2_install for uefi if we can't verify the
-            # uefi bits
-            utils.execute('chroot %(path)s /bin/sh -c '
-                          '"%(bin)s-install --removable"' %
-                          {'path': path, 'bin': binary_name},
-                          shell=True,
-                          env_variables={
-                              'PATH': path_variable
-                          })
+            try:
+                utils.execute('chroot %(path)s /bin/sh -c '
+                              '"%(bin)s-install"' %
+                              {'path': path, 'bin': binary_name},
+                              shell=True,
+                              env_variables={
+                                  'PATH': path_variable
+                              })
+            except processutils.ProcessExecutionError as e:
+                LOG.warning('Ignoring GRUB2 boot loader installation failure: '
+                            '%s.', e)
+            try:
+                # Also run grub-install with --removable, this installs grub to
+                # the EFI fallback path. Useful if the NVRAM wasn't written
+                # correctly, was reset or if testing with virt as libvirt
+                # resets the NVRAM on instance start.
+                # This operation is essentially a copy operation. Use of the
+                # --removable flag, per the grub-install source code changes
+                # the default file to be copied, destination file name, and
+                # prevents NVRAM from being updated.
+                # We only run grub2_install for uefi if we can't verify the
+                # uefi bits
+                utils.execute('chroot %(path)s /bin/sh -c '
+                              '"%(bin)s-install --removable"' %
+                              {'path': path, 'bin': binary_name},
+                              shell=True,
+                              env_variables={
+                                  'PATH': path_variable
+                              })
+            except processutils.ProcessExecutionError as e:
+                LOG.warning('Ignoring GRUB2 boot loader installation failure: '
+                            '%s.', e)
             utils.execute('umount', efi_partition_mount_point, attempts=3,
                           delay_on_retry=True)
             efi_mounted = False
